@@ -7,13 +7,15 @@ import json
 import re
 import multiprocessing
 
-__author__ = 'Yixin Chen & Haiyang Bian'
+__author__ = 'Yixin Chen & Haiyang Bian & Minsheng Hao'
 
 # global variables 
 df = pd.DataFrame()
 client_global = None
 _tablename = None
 
+# [0621]load tree relationship file
+df_tree = pd.read_csv("uhafOBO_0621.csv")
 
 ####### Preprocessing functions #######
 
@@ -111,7 +113,7 @@ def _forward2afterward(forward_seq):
 
 
 # transform an string into OTS BoolQuery to set searching condition in metadata
-def _seq2boolquery(metadata_condition):
+def _seq2boolquery(metadata_condition,include_children = True):
     ## [0601] transform uHAF_name to hcad_name
     metadata_condition = metadata_condition.replace("uHAF_name","hcad_name")
     # standardize string
@@ -126,7 +128,26 @@ def _seq2boolquery(metadata_condition):
         if word not in ['||','&&','!']:
             if "==" in word:
                 w_pair = word.split("==")
-                stack.append(tablestore.TermQuery(w_pair[0], w_pair[1]))
+                # [0621]
+                if w_pair[0] == "cell_type"  and include_children:
+                    # find all childs in the table
+                    stack_father = []
+                    stack_child = []
+                    stack_father.append("T cell")
+                    while not len(stack_father)==0:
+                        father = stack_father.pop()
+                        stack_child.append(father)
+                        childs = df_tree['child'][df_tree['father'] == father]
+                        for c in childs:
+                            stack_father.append(c)
+
+                    all_childs = np.unique(stack_child)
+                    child_queries = []
+                    for c in all_childs:
+                        child_queries.append(tablestore.TermQuery("cell_type",c))
+                    stack.append(tablestore.BoolQuery(should_queries=child_queries,minimum_should_match=1))
+                else:
+                    stack.append(tablestore.TermQuery(w_pair[0], w_pair[1]))
             elif "<>" in word:
                 w_pair = word.split("<>")
                 stack.append(tablestore.BoolQuery(must_not_queries=[tablestore.TermQuery(w_pair[0], w_pair[1])]))
@@ -497,12 +518,12 @@ def build_index():
         client_global.create_search_index(_tablename, "metadata", index_meta)
 
 
-def query_cells(metadata_conditions):
+def query_cells(metadata_conditions,include_children=False):
     """query cells and return the list of their primary keys
     
         This function is used to query cells satisfing the metadata conditions in the OTS table on the server. The conditions on the metadata should be a structured string which is a combination of several logical expressions. 
         
-        Users should remember that this function only enable them to query cell based on the conditions on the 17 metadata columns in the index while conditions on gene expression should be set in the get_columnsbycell function or get_columnsbycell_para function.
+        Users should remember that this function only enable them to query cell based on the conditions on the metadata columns in the index while conditions on gene expression should be set in the get_columnsbycell function or get_columnsbycell_para function.
     
     
     Parameters
@@ -531,7 +552,10 @@ def query_cells(metadata_conditions):
 
             'organ == Heart &&(cell_type == Fibrocyte || region <> atria)'
 
-
+    include_children: bool 
+        include_children is the parameter deciding if subtypes should be included in the query with condition on cell type. For example, when we query "cell_type == T cell" and set include_children as True, 
+        cells in the query result would contain T cell and the subtypes of T cells like CD4 T cell, CD8 T cell and so on; otherwise, only T cells would exist in the query result.
+    
     Returns
     -------
     list
@@ -543,30 +567,35 @@ def query_cells(metadata_conditions):
 
         The primary key list can be used for downstream data downloading or table updating.
     """
-    query = _seq2boolquery(metadata_conditions)
+    query = _seq2boolquery(metadata_conditions,include_children)
     # get primary keys
     results = []
     next_token = None
     total_cells = 0
-    rows, next_token, total_count, is_all_succeed = client_global.search(
+    # [0620] update for tablestore 5.2.1
+    search_result = client_global.search(
         _tablename, "metadata", 
         tablestore.SearchQuery(query, next_token=next_token, limit=100, get_total_count=True), 
         tablestore.ColumnsToGet(return_type=tablestore.ColumnReturnType.NONE)
     )
-    results += rows
-    total_cells += total_count
+    next_token = search_result.next_token
+    results += search_result.rows
+    total_cells += search_result.total_count
 
     while next_token:# if not finished
-        rows, next_token, total_count, is_all_succeed = client_global.search(
+        # [0620] update for tablestore 5.2.1
+        search_result = client_global.search(
         _tablename, "metadata", 
         tablestore.SearchQuery(query, next_token=next_token, limit=100, get_total_count=True), 
         tablestore.ColumnsToGet(return_type=tablestore.ColumnReturnType.NONE)  
         )
-        results += rows
-        total_cells += total_count
-    print("%d cells found" %total_count)
+        next_token = search_result.next_token
+        results += search_result.rows
+        total_cells += search_result.total_count
+    
     # build the list of the primary keys of the quried cells
-    rows_to_get = [x[0] for x in results]    
+    rows_to_get = [x[0] for x in results]
+    print("%d cells found" %len(results))    
     return rows_to_get
 
 
@@ -597,8 +626,8 @@ def _para_get_column(arg_bag):
     primary_key, columns_to_get, col_filter = arg_bag
     consumed, return_row, next_token= client_global.get_row(_tablename, primary_key,columns_to_get,column_filter=col_filter)
     ### [0601] avoid errors from NA in the row
-    if not return_row is None:
-        return [(x[0],x[1]) for x in return_row.primary_key] + [(x[0],x[1]) for x in return_row.attribute_columns]
+    return [(x[0],x[1]) for x in return_row.primary_key] + [(x[0],x[1]) for x in return_row.attribute_columns]
+    #if not return_row is None:
     #   return ([x[1] for x in return_row.primary_key] + [x[1] for x in return_row.attribute_columns])
 
 
@@ -794,10 +823,12 @@ def get_all_rows(cols_to_get = []):
         col_return_type = tablestore.ColumnReturnType.SPECIFIED
 
     while not all_rows or next_token:
-        rows, next_token, total_count, is_all_succeed = client_global.search(_tablename, 'metadata',
+        # [0620] update for tablestore 5.2.1
+        search_result = client_global.search(_tablename, 'metadata',
             tablestore.SearchQuery(query, next_token=next_token, limit=100, get_total_count=True),
             columns_to_get=tablestore.ColumnsToGet(cols_to_get, col_return_type))
-        all_rows.extend(rows)
+        next_token = search_result.next_token
+        all_rows.extend(search_result.rows)
 
     print("%d cells found" %len(all_rows))
 
